/*
 * uhub - A tiny ADC p2p connection hub
 * Copyright (C) 2007, Jan Vidar Krey
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 *
 */

#include "uhub.h"

void adc_cmd_incref(struct adc_command* msg)
{
	uhub_log(log_dump, "cmd: incref %p", msg);
	msg->references++;
}


void adc_cmd_free(struct adc_command* msg)
{
	msg->references--;
	
	if (msg->references < 0)
	{
		uhub_log(log_error, "cmd: free %p (DOUBLE FREE!)", msg);
		return;
	}
	
	if (msg->references == 0)
	{
		uhub_log(log_dump, "cmd: free %p (free)", msg);
		free(msg->cache);
		
		if (msg->feature_cast_include)
		{
			list_clear(msg->feature_cast_include, &free);
			list_destroy(msg->feature_cast_include);
		}
		
		if (msg->feature_cast_exclude)
		{
			list_clear(msg->feature_cast_exclude, &free);
			list_destroy(msg->feature_cast_exclude);
		}
	
		free(msg);
		msg = 0;
	}
	else
	{
		uhub_log(log_dump, "cmd: free %p (refcount: %d)", msg, (int) msg->references);
	}
}


struct adc_command* adc_cmd_copy(struct adc_command* cmd)
{
	char* tmp = 0;
	struct adc_command* copy = (struct adc_command*) malloc(sizeof(struct adc_command));
	if (!copy)
		return NULL; /* OOM */

	memcpy(copy, cmd, sizeof(struct adc_command));
	if (cmd->cache) copy->cache = strdup(cmd->cache);
	copy->references = 0;
	
	if (cmd->feature_cast_include)
	{
		copy->feature_cast_include = list_create();
		tmp = list_get_first(cmd->feature_cast_include);
		while (tmp)
		{
			list_append(copy->feature_cast_include, strdup(tmp));
			tmp = list_get_next(cmd->feature_cast_include);
		}
	}
	
	if (cmd->feature_cast_exclude)
	{
		copy->feature_cast_exclude = list_create();
		tmp = list_get_first(cmd->feature_cast_exclude);
		while (tmp)
		{
			list_append(copy->feature_cast_exclude, strdup(tmp));
			tmp = list_get_next(cmd->feature_cast_exclude);
		}
	}
	return copy;
}


struct adc_command* adc_cmd_parse(struct user* u, const char* line)
{
	struct adc_command* command = (struct adc_command*) malloc(sizeof(struct adc_command));
	char prefix = line[0];
	size_t length = strlen(line);
	size_t n = 0;
	char temp_sid[5];
	int ok = 1;
	struct linked_list* feature_cast_list;
	
	if (command == NULL)
		return NULL; /* OOM */

	memset(command, 0, sizeof(struct adc_command));
	command->length = length + 2;
	command->capacity = command->length;
	command->cache = (char*) malloc(command->length);
	
	if (command->cache == NULL)
	{
		free(command);
		return NULL; /* OOM */
	}
	
	memcpy(command->cache, line, command->length);
	command->cache[length - 0] = '\n';
	command->cache[length + 1] = '\0';
	command->cmd = FOURCC(line[0], line[1], line[2], line[3]);
	command->priority = 0;
	command->references = 0;
	command->arg_offset = 4;
	
	switch (prefix)
	{
		case 'I':
		case 'H':
			ok = (length > 3);
			command->arg_offset = 4;
			break;
			
		case 'B':
			ok = (length > 8 &&
					is_space(line[4]) &&
					is_valid_base32_char(line[5]) &&
					is_valid_base32_char(line[6]) &&
					is_valid_base32_char(line[7]) &&
					is_valid_base32_char(line[8]));
			
			if (!ok) break;
			
			temp_sid[0] = line[5];
			temp_sid[1] = line[6];
			temp_sid[2] = line[7];
			temp_sid[3] = line[8];
			temp_sid[4] = '\0';
			
			command->source = string_to_sid(temp_sid);
			ok = (command->source == u->sid);
			
			command->arg_offset = 9;
			break;
			
		case 'F':
			ok = (length > 8 &&
					is_space(line[4]) &&
					is_valid_base32_char(line[5]) &&
					is_valid_base32_char(line[6]) &&
					is_valid_base32_char(line[7]) &&
					is_valid_base32_char(line[8]));
			
			if (!ok) break;
			
			temp_sid[0] = line[5];
			temp_sid[1] = line[6];
			temp_sid[2] = line[7];
			temp_sid[3] = line[8];
			temp_sid[4] = '\0';
			
			command->source = string_to_sid(temp_sid);
			ok = (command->source == u->sid);
			
			/* Create feature cast lists */
			command->feature_cast_include = list_create();
			command->feature_cast_exclude = list_create();
			
			if (!command->feature_cast_include || !command->feature_cast_exclude)
			{
				list_destroy(command->feature_cast_include);
				list_destroy(command->feature_cast_exclude);
				free(command->cache);
				free(command);
				return NULL; /* OOM */
			}
			
			n = 10;
			while (line[n] == '+' || line[n] == '-')
			{
				if (line[n++] == '+')
					feature_cast_list = command->feature_cast_include;
				else
					feature_cast_list = command->feature_cast_exclude;
				
				temp_sid[0] = line[n++];
				temp_sid[1] = line[n++];
				temp_sid[2] = line[n++];
				temp_sid[3] = line[n++];
				temp_sid[4] = '\0';
				
				list_append(feature_cast_list, strdup(temp_sid));
			}
			
			command->arg_offset = n;
			
			if  (n == 9)
				ok = 0;
			
			break;
			
		case 'D':
		case 'E':
			ok = (length > 13 &&
					is_space(line[4]) &&
					is_valid_base32_char(line[5]) &&
					is_valid_base32_char(line[6]) &&
					is_valid_base32_char(line[7]) &&
					is_valid_base32_char(line[8]) &&
					is_space(line[9]) &&
					is_valid_base32_char(line[10]) &&
					is_valid_base32_char(line[11]) &&
					is_valid_base32_char(line[12]) &&
					is_valid_base32_char(line[13]));
			
			if (!ok) break;
			
			temp_sid[0] = line[5];
			temp_sid[1] = line[6];
			temp_sid[2] = line[7];
			temp_sid[3] = line[8];
			temp_sid[4] = '\0';
			
			command->source = string_to_sid(temp_sid);
			ok = (command->source == u->sid);
			
			temp_sid[0] = line[10];
			temp_sid[1] = line[11];
			temp_sid[2] = line[12];
			temp_sid[3] = line[13];
			temp_sid[4] = '\0';
			
			command->target = string_to_sid(temp_sid);
			
			command->arg_offset = 14;
			break;
			
		default:
			ok = 0;
	}
	
	if (!ok) {
		adc_cmd_free(command);
		return NULL;
	}
	
	/* Either 'n' is space, or end of line here! Anything else is bogus! */
	if (line[command->arg_offset] == ' ')
	{
		command->arg_offset++;
		if (line[command->arg_offset] == ' ') ok = 0;
	}
	else if (line[command->arg_offset] == '\0') ok = 1;
	else ok = 0;
	
	if (!ok)
	{
		adc_cmd_free(command);
		return NULL;
	}
	
	uhub_log(log_dump, "cmd: parsed %p", command);
	adc_cmd_incref(command);
	return command;
}


struct adc_command* adc_cmd_create(const char* line)
{
	return adc_cmd_parse(NULL, line);
}


int adc_cmd_remove_named_argument(struct adc_command* cmd, const char prefix_[2])
{
	char* start;
	char* end;
	size_t endlen;
	char prefix[4] = { ' ', prefix_[0], prefix_[1], '\0' };
	int found = 0;
	
	/* uhub_log(log_dump, "Removing argument %s from command", prefix); */
	start = strstr(&cmd->cache[cmd->arg_offset-1], prefix);
	while (start)
	{
		end = strchr(&start[1], ' ');
		if (end != NULL)
		{
			endlen = strlen(end);
			memmove(start, end, endlen);
			start[endlen] = '\0';
			found++;
		}
		else
		{
			start[0] = '\n';
			start[1] = '\0';
			found++;
			break;
		}
		start = strstr(&cmd->cache[cmd->arg_offset-1], prefix);
	}
	return found;
}


int adc_cmd_has_named_argument(struct adc_command* cmd, const char prefix_[2])
{
	int count = 0;
	char* start;
	char prefix[4] = { ' ', prefix_[0], prefix_[1], '\0' };

	uhub_log(log_trace, "Checking for argument %s", prefix);

	start = strstr(&cmd->cache[cmd->arg_offset-1], prefix);
	while (start)
	{
		count++;
		start = strstr(&start[1], prefix);
	}
	
	return count;
}


char* adc_cmd_get_named_argument(struct adc_command* cmd, const char prefix_[2])
{
	char* start;
	char* end;
	char* argument;
	size_t length;
	char prefix[4] = { ' ', prefix_[0], prefix_[1], '\0' };
	
	start = strstr(&cmd->cache[cmd->arg_offset-1], prefix);
	if (start == NULL)
		return NULL;
	
	start = &start[3];
	end = strchr(start, ' ');
	if (end == NULL)
		length = strlen(start);
	else
		length = &end[0] - &start[0];

	
	argument = malloc(length + 1);
	if (argument == NULL)
		return NULL;

	memcpy(argument, start, length);
	argument[length] = 0;
	return argument;
}


int adc_cmd_replace_named_argument(struct adc_command* cmd, const char prefix[2], const char* string)
{
	if (adc_cmd_has_named_argument(cmd, prefix))
	{
		adc_cmd_remove_named_argument(cmd, prefix);
	}
	
	if (adc_cmd_add_named_argument(cmd, prefix, string) == -1)
	{
		return -1;
	}
	return 0;
}


void adc_cmd_terminate_command(struct adc_command* cmd)
{
	char* data = cmd->cache;
	size_t len = strlen(data);
	
	if (len > 0 && data[len-1] != '\n')
	{
		if (len > cmd->capacity)
		{
			data = malloc(len + 2); /* FIXME: OOM! */
			strcpy(data, cmd->cache);
			free(cmd->cache);
			cmd->cache = data;
			cmd->capacity = len+1;
		}
		
		data[len+0] = '\n';
		data[len+1] = 0;
		cmd->length++;
	}
}


void adc_cmd_unterminate_command(struct adc_command* cmd)
{
	char* data = cmd->cache;
	size_t len = strlen(data);
	
	if (len > 0 && data[len-1] == '\n')
	{
		data[len-1] = 0;
		cmd->length--;
	}
}


int adc_cmd_add_named_argument(struct adc_command* cmd, const char prefix_[2], const char* string)
{
	char prefix[4] = { ' ', prefix_[0], prefix_[1], '\0' };
	char* data = 0;
	size_t len = 0;
	
	len = strlen(cmd->cache) + strlen(string) + 3;
	if (len > cmd->capacity)
	{
		data = malloc(len + 2);
		if (data == NULL)
		{
			return -1;
		}
		strcpy(data, cmd->cache);
		if (strlen(data) > 0 && data[strlen(data)-1] == '\n')
			data[strlen(data)-1] = 0;
		
		free(cmd->cache);
		cmd->cache = data;
		cmd->capacity = len;
		cmd->length = len;
	}
	else
	{
		data = cmd->cache;
		if (strlen(data) > 0 && data[strlen(data)-1] == '\n')
		{
			data[strlen(data)-1] = 0;
			cmd->length++;
		}
		
	}
	
	strcat(data, prefix);
	strcat(data, string);
	
	adc_cmd_terminate_command(cmd);
	
	return 0;
}


int adc_cmd_add_argument(struct adc_command* cmd, const char* string)
{
	char* data = 0;
	size_t len = 0;
	
	len = strlen(cmd->cache) + strlen(string);
	if (len > cmd->capacity)
	{
		data = malloc(len + 2);
		if (data == NULL)
		{
			return -1;
		}
		strcpy(data, cmd->cache);
		if (strlen(data) > 0 && data[strlen(data)-1] == '\n')
			data[strlen(data)-1] = 0;
		
		free(cmd->cache);
		cmd->cache = data;
		cmd->length = len;
		cmd->capacity = len;
	}
	else
	{
		data = cmd->cache;
		if (strlen(data) > 0 && data[strlen(data)-1] == '\n')
			data[strlen(data)-1] = 0;
	}
	
	strcat(data, " ");
	strcat(data, string);
	
	adc_cmd_terminate_command(cmd);
	return 0;
}


char* adc_cmd_get_argument(struct adc_command* cmd, int offset)
{
	char* start;
	char* end;
	char* argument;
	int count = 0;
	
	start = strchr(&cmd->cache[cmd->arg_offset-1], ' ');
	while (start)
	{
		end = strchr(&start[1], ' ');
		
		if (count == offset)
		{
			if (end)
			{
				argument = strndup(&start[1], (&end[0] - &start[1]));
			}
			else
			{
				argument = strdup(&start[1]);
				if (argument[strlen(argument)-1] == '\n')
					argument[strlen(argument)-1] = 0;
			}
			
			if (strlen(argument))
				return argument;
		}
		
		count++;
		start = end;
	}
	return 0;
}


int adc_cmd_escape_length(const char* str)
{
	int add = 0;
	int n = 0;
	for (; str[n]; n++)
		if (str[n] == ' ' || str[n] == '\n' || str[n] == '\\') add++;
	return n + add;
}


int adc_cmd_unescape_length(const char* str)
{
	int add = 0;
	int n = 0;
	int escape = 0;
	for (; str[n]; n++)
	{
		if (escape)
		{
			escape = 0;
		}
		else
		{
			if (str[n] == '\\')
			{
				escape = 1;
				add++;
			}
		}
	}
	return n - add;
}


char* adc_cmd_unescape(const char* string)
{
	char* new_string = malloc(adc_cmd_unescape_length(string)+1);
	char* ptr = (char*) new_string;
	char* str = (char*) string;
	int escaped = 0;
	
	while (*str)
	{
		if (escaped) {
			if (*str == 's')
				*ptr++ = ' ';
			else if (*str == '\\')
				*ptr++ = '\\';
			else if (*str == 'n')
				*ptr++ = '\n';
			else
				*ptr++ = *str;
			
			escaped = 0;
		} else {
			if (*str == '\\')
				escaped = 1;
			else
				*ptr++ = *str;
				
		}
		str++;
	}
	*ptr = 0;
	return new_string;
}


char* adc_cmd_escape(const char* string)
{
	char* str = malloc(adc_cmd_escape_length(string)+1);
	int n = 0;
	int i = 0;
	for (i = 0; i < strlen(string); i++)
	{
		switch (string[i]) {
			case '\\': /* fall through */
					str[n++] = '\\';
					str[n++] = '\\';
					break;
			case '\n':
					str[n++] = '\\';
					str[n++] = 'n';
					break;
			case ' ':
					str[n++] = '\\';
					str[n++] = 's';
					break;
			default:
					str[n++] = string[i];
					break;
		}
	}
	str[n] = '\0';
	return str;
}

