/*
 * uhub - A tiny ADC p2p connection hub
 * Copyright (C) 2007, Jan Vidar Krey
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 *
 */

#ifndef HAVE_uhub_COMMAND_H
#define HAVE_uhub_COMMAND_H

#include "uhub.h"


struct user;

struct adc_command
{
	fourcc_t cmd;
	sid_t source;
	sid_t target;
	char* cache;
	int length;
	int capacity;
	int arg_offset;
	int priority;
	int references;
	struct linked_list*  feature_cast_include;
	struct linked_list*  feature_cast_exclude;
};


/**
 * Increase the adc command reference counter
 */
extern void adc_cmd_incref(struct adc_command* msg);

/**
 * Decrease the reference counter, and free the memory when apropriate.
 */
extern void adc_cmd_free(struct adc_command* msg);

/**
 * Copy a command. Note: 'references' will be zero for the copied command.
 * @return a copy of cmd or NULL if not able to allocate memory.
 */
extern struct adc_command* adc_cmd_copy(struct adc_command* cmd);

/**
 * This will parse 'string' and return it as a adc_command struct, or
 * NULL if not able to allocate memory or 'string' does not contain
 * a valid ADC command.
 */
extern struct adc_command* adc_cmd_parse(struct user* u, const char* string);

/**
 * This will construct a adc_command based on 'string'.
 * Only to be used for server generated commands.
 */
extern struct adc_command* adc_cmd_create(const char* string);

/**
 * Remove a named argument from the command.
 *
 * @arg prefix a 2 character argument prefix
 * @return the number of named arguments removed.
 */
extern int adc_cmd_remove_named_argument(struct adc_command* cmd, const char prefix[2]);

/**
 * Count the number of arguments matching the given 2 character prefix.
 *
 * @arg prefix a 2 character argument prefix
 * @return the number of matching arguments
 */
extern int adc_cmd_has_named_argument(struct adc_command* cmd, const char prefix[2]);

/**
 * Returns a named arguments based on the 2 character prefix.
 * If multiple matching arguments exists, only the first one will be returned
 * by this function.
 *
 * @arg prefix a 2 character argument prefix
 * @return the argument or NULL if OOM/not found.
 */
extern char* adc_cmd_get_named_argument(struct adc_command* cmd, const char prefix[2]);

/**
 * Returns the argument on the offset position in the command.
 * If offset is invalid NULL is returned.
 * @return the argument or NULL if OOM/not found.
 */
extern char* adc_cmd_get_argument(struct adc_command* cmd, int offset);

/**
 * Replace a named argument in the command.
 * This will remove any matching arguments (multiple, or none),
 * then add 'string' as an argument using the given prefix.
 *
 * @arg prefix a 2 character argument prefix
 * @arg string must be escaped (see adc_cmd_escape).
 * @return  0 if successful, or -1 if an error occured.
 */
extern int adc_cmd_replace_named_argument(struct adc_command* cmd, const char prefix[2], const char* string);

/**
 * Append an argument
 *
 * @arg string must be escaped (see adc_cmd_escape).
 * @return  0 if successful, or -1 if an error occured (out of memory).
 */
extern int adc_cmd_add_argument(struct adc_command* cmd, const char* string);

/**
 * Append a named argument
 *
 * @arg prefix a 2 character argument prefix
 * @arg string must be escaped (see adc_cmd_escape).
 * @return  0 if successful, or -1 if an error occured (out of memory).
 */
extern int adc_cmd_add_named_argument(struct adc_command* cmd, const char prefix[2], const char* string);

/**
 * Convert a ADC command escaped string to a regular string.
 * @return string or NULL if out of memory
 */
extern char* adc_cmd_unescape(const char* string);

/**
 * Convert a string to a ADC command escaped string.
 * @return adc command escaped string or NULL if out of memory.
 */
extern char* adc_cmd_escape(const char* string);

/**
 * This will ensure a newline is at the end of the command.
 */
void adc_cmd_terminate_command(struct adc_command* cmd);

/**
 * This will remove any newline from the end of the command
 */
void adc_cmd_unterminate_command(struct adc_command* cmd);

#endif /* HAVE_uhub_COMMAND_H */
